from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.cstruct import cstruct

if TYPE_CHECKING:
    from enum import Flag, IntEnum

elf_32_def = """
typedef uint32      Elf32_Addr;
typedef uint16      Elf32_Half;
typedef uint32      Elf32_Off;
typedef int32       Elf32_Sword;
typedef uint32      Elf32_Word;

typedef struct {
    Elf32_Addr      r_offset;
    Elf32_Word      r_info;
} Rel;

typedef struct {
    Elf32_Addr      r_offset;
    Elf32_Word      r_info;
    Elf32_Sword     r_addend;
} Rela;

typedef struct {
    Elf32_Word      st_name;
    Elf32_Addr      st_value;
    Elf32_Word      st_size;
    char            st_info;
    char            st_other;
    Elf32_Half      st_shndx;
} Sym;

typedef struct {
    char            e_ident[EI_NIDENT];
    Elf_Type        e_type;
    Elf32_Half      e_machine;
    Elf32_Word      e_version;
    Elf32_Addr      e_entry;             /* Entry point */
    Elf32_Off       e_phoff;
    Elf32_Off       e_shoff;
    Elf32_Word      e_flags;
    Elf32_Half      e_ehsize;
    Elf32_Half      e_phentsize;
    Elf32_Half      e_phnum;
    Elf32_Half      e_shentsize;
    Elf32_Half      e_shnum;
    Elf32_Half      e_shstrndx;
} Ehdr;

typedef struct {
    PT              p_type;
    Elf32_Off       p_offset;
    Elf32_Addr      p_vaddr;
    Elf32_Addr      p_paddr;
    Elf32_Word      p_filesz;
    Elf32_Word      p_memsz;
    PF              p_flags;
    Elf32_Word      p_align;
} Phdr;

typedef struct {
    Elf32_Word      sh_name;
    SHT             sh_type;
    Elf32_Word      sh_flags;
    Elf32_Addr      sh_addr;
    Elf32_Off       sh_offset;
    Elf32_Word      sh_size;
    Elf32_Word      sh_link;
    Elf32_Word      sh_info;
    Elf32_Word      sh_addralign;
    Elf32_Word      sh_entsize;
} Shdr;

/* Note header in a PT_NOTE section */
typedef struct {
    Elf32_Word      n_namesz;            /* Name size */
    Elf32_Word      n_descsz;            /* Content size */
    Elf32_Word      n_type;              /* Content type */
} Nhdr;
"""

elf_64_def = """
typedef uint64      Elf64_Addr;
typedef uint16      Elf64_Half;
typedef int16       Elf64_SHalf;
typedef uint64      Elf64_Off;
typedef int32       Elf64_Sword;
typedef uint32      Elf64_Word;
typedef uint64      Elf64_Xword;
typedef int64       Elf64_Sxword;

typedef struct {
    Elf64_Addr      r_offset;            /* Location at which to apply the action */
    Elf64_Xword     r_info;              /* index and type of relocation */
} Rel;

typedef struct {
    Elf64_Addr      r_offset;            /* Location at which to apply the action */
    Elf64_Xword     r_info;              /* index and type of relocation */
    Elf64_Sxword    r_addend;            /* Constant addend used to compute value */
} Rela;

typedef struct {
    Elf64_Word      st_name;             /* Symbol name, index in string tbl */
    char            st_info;             /* Type and binding attributes */
    char            st_other;            /* No defined meaning, 0 */
    Elf64_Half      st_shndx;            /* Associated section index */
    Elf64_Addr      st_value;            /* Value of the symbol */
    Elf64_Xword     st_size;             /* Associated symbol size */
} Sym;

typedef struct {
    char            e_ident[EI_NIDENT];  /* ELF "magic number" */
    Elf_Type        e_type;
    Elf64_Half      e_machine;
    Elf64_Word      e_version;
    Elf64_Addr      e_entry;             /* Entry point virtual address */
    Elf64_Off       e_phoff;             /* Program header table file offset */
    Elf64_Off       e_shoff;             /* Section header table file offset */
    Elf64_Word      e_flags;
    Elf64_Half      e_ehsize;
    Elf64_Half      e_phentsize;
    Elf64_Half      e_phnum;
    Elf64_Half      e_shentsize;
    Elf64_Half      e_shnum;
    Elf64_Half      e_shstrndx;
} Ehdr;

typedef struct {
    PT              p_type;
    PF              p_flags;
    Elf64_Off       p_offset;             /* Segment file offset */
    Elf64_Addr      p_vaddr;              /* Segment virtual address */
    Elf64_Addr      p_paddr;              /* Segment physical address */
    Elf64_Xword     p_filesz;             /* Segment size in file */
    Elf64_Xword     p_memsz;              /* Segment size in memory */
    Elf64_Xword     p_align;              /* Segment alignment, file & memory */
} Phdr;

typedef struct {
    Elf64_Word      sh_name;              /* Section name, index in string tbl */
    SHT             sh_type;              /* Type of section */
    Elf64_Xword     sh_flags;             /* Miscellaneous section attributes */
    Elf64_Addr      sh_addr;              /* Section virtual addr at execution */
    Elf64_Off       sh_offset;            /* Section file offset */
    Elf64_Xword     sh_size;              /* Size of section in bytes */
    Elf64_Word      sh_link;              /* Index of another section */
    Elf64_Word      sh_info;              /* Additional section information */
    Elf64_Xword     sh_addralign;         /* Section alignment */
    Elf64_Xword     sh_entsize;           /* Entry size if section holds table */
} Shdr;

/* Note header in a PT_NOTE section */
typedef struct {
    Elf64_Word      n_namesz;             /* Name size */
    Elf64_Word      n_descsz;             /* Content size */
    Elf64_Word      n_type;               /* Content type */
} Nhdr;
"""

elf_def = """
#define PT_LOOS         0x60000000        /* OS-specific */
#define PT_HIOS         0x6fffffff        /* OS-specific */
#define PT_LOPROC       0x70000000
#define PT_HIPROC       0x7fffffff
#define PT_GNU_EH_FRAME 0x6474e550

enum PT : uint32 {
    NULL                = 0,
    LOAD                = 1,
    DYNAMIC             = 2,
    INTERP              = 3,
    NOTE                = 4,
    SHLIB               = 5,
    PHDR                = 6,
    TLS                 = 7
};

#define PT_GNU_STACK    (PT_LOOS + 0x474e551)

#define PN_XNUM         0xffff

/* These constants define the different elf file types */
#define ET_LOPROC       0xff00
#define ET_HIPROC       0xffff

enum Elf_Type : uint16 {
    ET_NONE             = 0, // Unkown Type
    ET_REL              = 1, // Relocatable File
    ET_EXEC             = 2, // Executable File
    ET_DYN              = 3,
    ET_CORE             = 4
};

/* This is the info that is needed to parse the dynamic section of the file */
#define DT_NULL         0
#define DT_NEEDED       1
#define DT_PLTRELSZ     2
#define DT_PLTGOT       3
#define DT_HASH         4
#define DT_STRTAB       5
#define DT_SYMTAB       6
#define DT_RELA         7
#define DT_RELASZ       8
#define DT_RELAENT      9
#define DT_STRSZ        10
#define DT_SYMENT       11
#define DT_INIT         12
#define DT_FINI         13
#define DT_SONAME       14
#define DT_RPATH        15
#define DT_SYMBOLIC     16
#define DT_REL          17
#define DT_RELSZ        18
#define DT_RELENT       19
#define DT_PLTREL       20
#define DT_DEBUG        21
#define DT_TEXTREL      22
#define DT_JMPREL       23
#define DT_ENCODING     32
#define OLD_DT_LOOS     0x60000000
#define DT_LOOS         0x6000000d
#define DT_HIOS         0x6ffff000
#define DT_VALRNGLO     0x6ffffd00
#define DT_VALRNGHI     0x6ffffdff
#define DT_ADDRRNGLO    0x6ffffe00
#define DT_ADDRRNGHI    0x6ffffeff
#define DT_VERSYM       0x6ffffff0
#define DT_RELACOUNT    0x6ffffff9
#define DT_RELCOUNT     0x6ffffffa
#define DT_FLAGS_1      0x6ffffffb
#define DT_VERDEF       0x6ffffffc
#define DT_VERDEFNUM    0x6ffffffd
#define DT_VERNEED      0x6ffffffe
#define DT_VERNEEDNUM   0x6fffffff
#define OLD_DT_HIOS     0x6fffffff
#define DT_LOPROC       0x70000000
#define DT_HIPROC       0x7fffffff

/* This info is needed when parsing the symbol table */
enum STB {
    LOCAL               = 0, /* Local scope */
    GLOBAL              = 1, /* Global scope */
    WEAK                = 2  /* Weak, (ie. __attribute__((weak))) */
};

enum STT {
    NOTYPE              = 0, // No type
    OBJECT              = 1, // Variables, arrays, etc.
    FUNC                = 2, // Methods or functions
    SECTION             = 3,
    FILE                = 4,
    COMMON              = 5,
    TLS                 = 6
};

enum STV {
    DEFAULT              = 0,
    INTERNAL             = 1,
    HIDDEN               = 2,
    PROTECTED            = 3
};

/* These constants define the permissions on sections in the program
   header, p_flags. */
flag PF : uint32 {
    R = 0x4,
    W = 0x2,
    X = 0x1
};

/* sh_type */
enum SHT : uint32 {
    NULL                = 0,   /* Null section */
    PROGBITS            = 1,   /* Program information */
    SYMTAB              = 2,   /* Symbol table */
    STRTAB              = 3,   /* String table */
    RELA                = 4,   /* Relocation (w/ addend) */
    HASH                = 5,
    DYNAMIC             = 6,
    NOTE                = 7,
    NOBITS              = 8,   /* Not present in file */
    REL                 = 9,   /* Relocation (no addend) */
    SHLIB               = 10,
    DYNSYM              = 11,
    NUM                 = 12
};

/* sh_flags */
enum SHN : uint16 {
    UNDEF               = 0x0000,
    LOPROC              = 0xff00,
    HIPROC              = 0xff1f,
    LIVEPATCH           = 0xff20,
    ABS                 = 0xfff1
    COMMON              = 0xfff2
    HIRESERVE           = 0xffff
};


#define EI_NIDENT       16                /* Size of e_ident[] */
#define EI_MAG0         0                 /* e_ident[] indexes */
#define EI_MAG1         1
#define EI_MAG2         2
#define EI_MAG3         3
#define EI_CLASS        4
#define EI_DATA         5                 /* Defines endianness */
#define EI_VERSION      6
#define EI_OSABI        7
#define EI_PAD          8

#define ELFMAG0         0x7f              /* EI_MAG */
#define ELFMAG1         'E'
#define ELFMAG2         'L'
#define ELFMAG3         'F'
#define ELFMAG          b"\x7fELF"
#define SELFMAG         4

#define ELFCLASSNONE    0                 /* EI_CLASS */
#define ELFCLASS32      1
#define ELFCLASS64      2
#define ELFCLASSNUM     3

#define ELFDATANONE     0                 /* e_ident[EI_DATA] */
#define ELFDATA2LSB     1
#define ELFDATA2MSB     2

#define EV_NONE         0                 /* e_version, EI_VERSION */
#define EV_CURRENT      1
#define EV_NUM          2

#define ELFOSABI_NONE   0
#define ELFOSABI_LINUX  3

/*
 * Notes used in ET_CORE. Architectures export some of the arch register sets
 * using the corresponding note types via the PTRACE_GETREGSET and
 * PTRACE_SETREGSET requests.
 */
#define NT_PRSTATUS     1
#define NT_PRFPREG      2
#define NT_PRPSINFO     3
#define NT_TASKSTRUCT   4
#define NT_AUXV         6

#define NT_SIGINFO          0x53494749
#define NT_FILE             0x46494c45
#define NT_PRXFPREG         0x46e62b7f  /* copied from gdb5.1/include/elf/common.h */
#define NT_PPC_VMX          0x100       /* PowerPC Altivec/VMX registers */
#define NT_PPC_SPE          0x101       /* PowerPC SPE/EVR registers */
#define NT_PPC_VSX          0x102       /* PowerPC VSX registers */
#define NT_PPC_TAR          0x103       /* Target Address Register */
#define NT_PPC_PPR          0x104       /* Program Priority Register */
#define NT_PPC_DSCR         0x105       /* Data Stream Control Register */
#define NT_PPC_EBB          0x106       /* Event Based Branch Registers */
#define NT_PPC_PMU          0x107       /* Performance Monitor Registers */
#define NT_PPC_TM_CGPR      0x108       /* TM checkpointed GPR Registers */
#define NT_PPC_TM_CFPR      0x109       /* TM checkpointed FPR Registers */
#define NT_PPC_TM_CVMX      0x10a       /* TM checkpointed VMX Registers */
#define NT_PPC_TM_CVSX      0x10b       /* TM checkpointed VSX Registers */
#define NT_PPC_TM_SPR       0x10c       /* TM Special Purpose Registers */
#define NT_PPC_TM_CTAR      0x10d       /* TM checkpointed Target Address Register */
#define NT_PPC_TM_CPPR      0x10e       /* TM checkpointed Program Priority Register */
#define NT_PPC_TM_CDSCR     0x10f       /* TM checkpointed Data Stream Control Register */
#define NT_PPC_PKEY         0x110       /* Memory Protection Keys registers */
#define NT_386_TLS          0x200       /* i386 TLS slots (struct user_desc) */
#define NT_386_IOPERM       0x201       /* x86 io permission bitmap (1=deny) */
#define NT_X86_XSTATE       0x202       /* x86 extended state using xsave */
#define NT_S390_HIGH_GPRS   0x300       /* s390 upper register halves */
#define NT_S390_TIMER       0x301       /* s390 timer register */
#define NT_S390_TODCMP      0x302       /* s390 TOD clock comparator register */
#define NT_S390_TODPREG     0x303       /* s390 TOD programmable register */
#define NT_S390_CTRS        0x304       /* s390 control registers */
#define NT_S390_PREFIX      0x305       /* s390 prefix register */
#define NT_S390_LAST_BREAK  0x306       /* s390 breaking event address */
#define NT_S390_SYSTEM_CALL 0x307       /* s390 system call restart data */
#define NT_S390_TDB         0x308       /* s390 transaction diagnostic block */
#define NT_S390_VXRS_LOW    0x309       /* s390 vector registers 0-15 upper half */
#define NT_S390_VXRS_HIGH   0x30a       /* s390 vector registers 16-31 */
#define NT_S390_GS_CB       0x30b       /* s390 guarded storage registers */
#define NT_S390_GS_BC       0x30c       /* s390 guarded storage broadcast control block */
#define NT_S390_RI_CB       0x30d       /* s390 runtime instrumentation */
#define NT_ARM_VFP          0x400       /* ARM VFP/NEON registers */
#define NT_ARM_TLS          0x401       /* ARM TLS register */
#define NT_ARM_HW_BREAK     0x402       /* ARM hardware breakpoint registers */
#define NT_ARM_HW_WATCH     0x403       /* ARM hardware watchpoint registers */
#define NT_ARM_SYSTEM_CALL  0x404       /* ARM system call number */
#define NT_ARM_SVE          0x405       /* ARM Scalable Vector Extension registers */
#define NT_ARC_V2           0x600       /* ARCv2 accumulator/extra registers */
#define NT_VMCOREDD         0x700       /* Vmcore Device Dump Note */
#define NT_MIPS_DSP         0x800       /* MIPS DSP ASE registers */
#define NT_MIPS_FP_MODE     0x801       /* MIPS floating-point mode */

"""


def copy_cstruct(src_struct: cstruct) -> cstruct:
    dst_struct = cstruct()
    dst_struct.consts = src_struct.consts.copy()
    dst_struct.typedefs = src_struct.typedefs.copy()
    dst_struct.lookups = src_struct.lookups.copy()
    return dst_struct


c_common_elf = cstruct().load(elf_def)
c_elf_32 = copy_cstruct(c_common_elf).load(elf_32_def)
c_elf_64 = copy_cstruct(c_common_elf).load(elf_64_def)

PT: IntEnum = c_common_elf.PT
Elf_Type: IntEnum = c_common_elf.Elf_Type
STB: IntEnum = c_common_elf.STB
STT: IntEnum = c_common_elf.STT
STV: IntEnum = c_common_elf.STV
PF: Flag = c_common_elf.PF
SHN: IntEnum = c_common_elf.SHN
SHT: IntEnum = c_common_elf.SHT



from __future__ import annotations

import io
from functools import cached_property, lru_cache
from operator import itemgetter
from typing import TYPE_CHECKING, BinaryIO, Generic, TypeVar

from dissect.executable.elf.c_elf import (
    SHN,
    SHT,
    STB,
    STT,
    STV,
    Elf_Type,
    c_common_elf,
    c_elf_32,
    c_elf_64,
)
from dissect.executable.exception import InvalidSignatureError

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator

    from dissect.cstruct import cstruct


class ELF:
    def __init__(self, fh: BinaryIO):
        self.fh = fh

        fh.seek(0)
        self.e_ident = fh.read(0x10)

        if self.e_ident[:4] != c_common_elf.ELFMAG:
            raise InvalidSignatureError(
                f"Invalid ELF header magic, expected {c_common_elf.ELFMAG!r}, got {self.e_ident[:4]!r}"
            )

        c_elf_version = c_elf_32
        if self.e_ident[c_common_elf.EI_CLASS] == c_common_elf.ELFCLASS64:
            c_elf_version = c_elf_64
        self.c_elf = c_elf_version

        # Determine endianess interpretation of bytes. This matters after the first 16 bytes.
        is_little = self.e_ident[c_common_elf.EI_DATA] == c_common_elf.ELFDATA2LSB
        self.c_elf.endian = "<" if is_little else ">"

        fh.seek(0)
        self.header = self.c_elf.Ehdr(fh)

        self.segments = SegmentTable.from_elf(self)
        self.sections = SectionTable.from_elf(self)
        self.symbol_tables: list[SymbolTable] = self.sections.by_type([SHT.SYMTAB, SHT.DYNSYM])

    def __repr__(self) -> str:
        return str(self.header)

    def dump(self) -> bytes:
        output_data = [
            self.segments.dump_table(),
            self.sections.dump_table(),
            *self.segments.dump_data(),
            *self.sections.dump_data(),
        ]
        output_data.sort(key=itemgetter(0))

        result = []
        output_size = 0
        for offset, output_bytes in output_data:
            output_offset = offset - output_size

            buf = None
            relative_offset = output_offset + len(output_bytes)
            if output_offset < 0 and relative_offset > 0:
                buf = output_bytes[abs(output_offset) :]
            elif output_offset >= 0:
                buf = (b"\x00" * output_offset) + output_bytes

            if buf is not None:
                result.append(buf)
                output_size += len(buf)

        return b"".join(result)

    @property
    def dynamic(self) -> bool:
        return self.header.e_type == Elf_Type.ET_DYN


T = TypeVar("T")


class Table(Generic[T]):
    def __init__(self, num: int) -> None:
        self.num = num
        self.items: list[T] = [None] * num

    def __iter__(self) -> Iterator[T]:
        for idx in range(self.num):
            yield self[idx]

    def __getitem__(self, idx: int) -> T:
        if self.items[idx] is None:
            self.items[idx] = self._create_item(idx)
        return self.items[idx]

    def _create_item(self, idx: int) -> T:
        raise NotImplementedError

    def find(self, condition: Callable[[T], bool], **kwargs) -> list[T]:
        return [item for item in self if condition(item, **kwargs)]


class Section:
    def __init__(self, fh: BinaryIO, idx: int | None = None, c_elf: cstruct = c_elf_64):
        self.fh = fh
        self.idx = idx

        self.c_elf = c_elf
        self.header = self.c_elf.Shdr(self.fh)
        self.type = self.header.sh_type
        self.entry_size = self.header.sh_entsize
        self.alignment = self.header.sh_addralign
        self.offset = self.header.sh_offset
        self.size = self.header.sh_size

        self._name = None
        self._link = None

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__} idx={self.idx} name={self.name!r} type={self.type}"
            f" offset=0x{self.offset:x} size=0x{self.size:x}>"
        )

    def _set_name(self, table: StringTable) -> None:
        if self.header.sh_name != SHN.UNDEF:
            self._name = table[self.header.sh_name]

    def _set_link(self, table: SectionTable) -> None:
        if self.header.sh_link != SHN.UNDEF:
            self._link = table[self.header.sh_link]

    @classmethod
    def from_section_table(cls, table: SectionTable, idx: int) -> Section:
        result = cls(table.fh, idx=idx, c_elf=table.c_elf)
        result._set_link(table)

        if sh_strtab := (result if idx == table._sh_strtab_idx else table._sh_strtab):
            result._set_name(sh_strtab)

        return result

    @property
    def name(self) -> str | None:
        return self._name

    def is_related(self, segment: Segment) -> bool:
        return segment.is_related(self)

    @property
    def link(self) -> Section | None:
        return self._link

    @cached_property
    def data(self) -> bytes:
        self.fh.seek(self.offset)
        return self.fh.read(self.size)


class SectionTable(Table[Section]):
    def __init__(
        self,
        fh: BinaryIO,
        offset: int,
        num: int,
        size: int,
        sh_strtab_idx: int | None = None,
        c_elf: cstruct = c_elf_64,
    ):
        super().__init__(num)
        self.fh = fh
        self.offset = offset
        self.size = size
        self.c_elf = c_elf

        self._sh_strtab_idx = sh_strtab_idx
        self._sh_strtab = None

        if sh_strtab_idx:
            self._sh_strtab: StringTable = self[sh_strtab_idx]

    def __repr__(self) -> str:
        return f"<SectionTable offset=0x{self.offset:x} size=0x{self.size:x}>"

    def _create_item(self, idx: int) -> Section:
        self.fh.seek(self.offset + self.size * idx)
        _, section_type = self.c_elf.uint32[2](self.fh.read(8))
        self.fh.seek(-8, io.SEEK_CUR)

        return_class = Section
        if section_type == SHT.STRTAB:
            return_class = StringTable
        if section_type in [SHT.DYNSYM, SHT.SYMTAB]:
            return_class = SymbolTable

        return return_class.from_section_table(self, idx)

    @classmethod
    def from_elf(cls, elf: ELF) -> SectionTable:
        return cls(
            elf.fh,
            elf.header.e_shoff,
            elf.header.e_shnum,
            elf.header.e_shentsize,
            elf.header.e_shstrndx,
            elf.c_elf,
        )

    def by_type(self, section_types: list[int] | int) -> list[Section]:
        types = section_types
        if not isinstance(section_types, list):
            types = [types]

        return self.find(lambda x: x.type in types)

    def related_sections(self, segment: Segment) -> list[Section]:
        return self.find(lambda x: x.is_related(segment))

    def by_name(self, name: str) -> list[Section]:
        return self.find(lambda x: name == x.name)

    def dump_table(self) -> tuple[int, bytes]:
        buf = bytearray()
        return self.offset, buf.join([x.header.dumps() for x in self])

    def dump_data(self) -> list[tuple[int, bytes]]:
        return [(x.offset, x.data) for x in self]


class Segment:
    def __init__(self, fh: BinaryIO, idx: int | None = None, c_elf: cstruct = c_elf_64):
        self.fh = fh
        self.idx = idx
        self.c_elf = c_elf

        self.header = c_elf.Phdr(fh)
        self.type = self.header.p_type
        self.flags = self.header.p_flags
        self.virtual_address = self.header.p_vaddr
        self.physical_address = self.header.p_paddr
        self.memory_size = self.header.p_memsz
        self.alignment = self.header.p_align
        self.offset = self.header.p_offset
        self.size = self.header.p_filesz

        self._data = b""
        self.patched = False

    def __repr__(self) -> str:
        return repr(self.header)

    @classmethod
    def from_segment_table(cls, table: SegmentTable, idx: int | None = None) -> Segment:
        fh = table.fh
        return cls(fh, idx, table.c_elf)

    @property
    def end(self) -> int:
        return self.offset + self.size

    def is_related(self, section: Section) -> bool:
        return self.offset <= section.offset < self.end

    @property
    def data(self) -> bytes:
        if not self._data:
            self.fh.seek(self.offset)
            self._data = self.fh.read(self.size)
        return self._data

    def _alignment_padding(self, data_length: int) -> bytes:
        padding = 0
        if self.header.p_align > 1:
            padding = data_length % self.header.p_align
        return b"\x00" * padding

    def patch(self, new_data: bytes) -> None:
        self.patched = True
        self._data = new_data + self._alignment_padding(len(new_data))
        self.header.p_filesz = len(self._data)


class SegmentTable(Table[Segment]):
    def __init__(self, fh: BinaryIO, offset: int, entries: int, size: int, c_elf: cstruct = c_elf_64):
        super().__init__(entries)
        self.fh = fh
        self.offset = offset
        self.size = size
        self.c_elf = c_elf

    def __repr__(self) -> str:
        return f"<SegmentTable offset=0x{self.offset:x} size=0x{self.size:x}>"

    def _create_item(self, idx: int) -> Segment:
        self.fh.seek(self.offset + self.size * idx)
        return Segment.from_segment_table(self, idx)

    @classmethod
    def from_elf(cls, elf: ELF) -> SegmentTable:
        header = elf.header
        offset = header.e_phoff
        entries = header.e_phnum
        size = header.e_phentsize
        return cls(fh=elf.fh, offset=offset, entries=entries, size=size, c_elf=elf.c_elf)

    def related_segments(self, section: Section) -> list[Segment]:
        return self.find(lambda x: x.is_related(section))

    def by_type(self, segment_types: list[int] | int) -> list[Segment]:
        types = segment_types
        if not isinstance(segment_types, list):
            types = [types]

        return self.find(lambda x: x.type in types)

    def dump_data(self) -> list[tuple[int, bytearray]]:
        return [(x.offset, x.data) for x in self]

    def dump_table(self) -> tuple[int, bytearray]:
        buf = bytearray()
        return self.offset, buf.join([x.header.dumps() for x in self])


class StringTable(Section):
    def __init__(self, fh: BinaryIO, idx: int | None = None, c_elf: cstruct = c_elf_64):
        super().__init__(fh, idx, c_elf)

        self._get_string = lru_cache(256)(self._get_string)

    def __getitem__(self, offset: int) -> str:
        return self._get_string(offset)

    def _get_string(self, index: int) -> str:
        if index > self.size or index == SHN.UNDEF:
            return None
        return self.c_elf.char[None](self.data[index:]).decode("utf8")


class Symbol:
    def __init__(self, fh: BinaryIO, idx: int | None = None, c_elf: cstruct = c_elf_64):
        self.symbol = c_elf.Sym(fh)
        self.idx = idx
        self.c_elf = c_elf

        endianness = "little" if c_elf.endian == "<" else "big"

        info = int.from_bytes(self.symbol.st_info, endianness)
        self.bind = STB(info >> 4)
        self.type = STT(info & 0xF)
        self.size = self.symbol.st_size

        other = int.from_bytes(self.symbol.st_other, endianness)
        self.visibility = STV(other & 0x3)

        self._name = None

    def __repr__(self) -> str:
        return (
            f"<Symbol idx={self.idx} value=0x{self.value:x} size={self.size} type={self.type} bind={self.bind}"
            f" visibility={self.visibility} shndex={self.symbol.st_shndx} name={self.name}>"
        )

    def _set_name(self, table: StringTable) -> None:
        self._name = table[self.symbol.st_name]

    @classmethod
    def from_symbol_table(cls, table: SymbolTable, idx: int) -> Symbol:
        offset = idx * table.entry_size
        data = table.data[offset : offset + table.entry_size]
        output = cls(io.BytesIO(data), idx, table.c_elf)
        output._set_name(table.link)
        return output

    @property
    def name(self) -> str:
        return self._name

    @property
    def value(self) -> int:
        return 0 if self.symbol.st_shndx == SHN.UNDEF else self.symbol.st_value

    def value_based_on_shndx(self, table: SectionTable) -> int:
        symloc = self.symbol.st_shndx
        value = self.value
        if symloc not in [SHN.UNDEF, SHN.ABS]:
            value += table[symloc].offset
        return value


class SymbolTable(Section, Table[Symbol]):
    def __init__(self, fh: BinaryIO, idx: int | None = None, c_elf: cstruct = c_elf_64):
        # Initializes Section info
        Section.__init__(self, fh, idx, c_elf)
        count = self.size // self.entry_size
        # Initializes Table info
        Table.__init__(self, count)

    def _create_item(self, idx: int) -> Symbol:
        return Symbol.from_symbol_table(self, idx)



from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING

from dissect.executable.pe.c_pe import c_pe
from dissect.executable.pe.directory.base import DataDirectory

if TYPE_CHECKING:
    from collections.abc import Iterator


class ExceptionDirectory(DataDirectory):
    """The exception directory of a PE file.

    Currently only shows the raw exception entries.
    """

    def __repr__(self) -> str:
        return f"<ExceptionDirectory entries={len(self.entries)}>"

    def __len__(self) -> int:
        return len(self.entries)

    def __iter__(
        self,
    ) -> Iterator[
        c_pe.IMAGE_RUNTIME_FUNCTION_ENTRY
        | c_pe.IMAGE_ARM_RUNTIME_FUNCTION_ENTRY
        | c_pe.IMAGE_ARM64_RUNTIME_FUNCTION_ENTRY
        | c_pe.IMAGE_ALPHA_RUNTIME_FUNCTION_ENTRY
        | c_pe.IMAGE_ALPHA64_RUNTIME_FUNCTION_ENTRY
        | c_pe.IMAGE_CE_RUNTIME_FUNCTION_ENTRY
        | c_pe.IMAGE_MIPS_RUNTIME_FUNCTION_ENTRY
    ]:
        return iter(self.entries)

    def __getitem__(
        self, idx: int
    ) -> (
        c_pe.IMAGE_RUNTIME_FUNCTION_ENTRY
        | c_pe.IMAGE_ARM_RUNTIME_FUNCTION_ENTRY
        | c_pe.IMAGE_ARM64_RUNTIME_FUNCTION_ENTRY
        | c_pe.IMAGE_ALPHA_RUNTIME_FUNCTION_ENTRY
        | c_pe.IMAGE_ALPHA64_RUNTIME_FUNCTION_ENTRY
        | c_pe.IMAGE_CE_RUNTIME_FUNCTION_ENTRY
        | c_pe.IMAGE_MIPS_RUNTIME_FUNCTION_ENTRY
    ):
        return self.entries[idx]

    @cached_property
    def entries(
        self,
    ) -> list[
        c_pe.IMAGE_RUNTIME_FUNCTION_ENTRY
        | c_pe.IMAGE_ARM_RUNTIME_FUNCTION_ENTRY
        | c_pe.IMAGE_ARM64_RUNTIME_FUNCTION_ENTRY
        | c_pe.IMAGE_ALPHA_RUNTIME_FUNCTION_ENTRY
        | c_pe.IMAGE_ALPHA64_RUNTIME_FUNCTION_ENTRY
        | c_pe.IMAGE_CE_RUNTIME_FUNCTION_ENTRY
        | c_pe.IMAGE_MIPS_RUNTIME_FUNCTION_ENTRY
    ]:
        """List of exception entries."""
        self.pe.vfh.seek(self.address)

        machine = self.pe.machine
        if machine in (c_pe.IMAGE_FILE_MACHINE.ARM, c_pe.IMAGE_FILE_MACHINE.THUMB, c_pe.IMAGE_FILE_MACHINE.ARMNT):
            ctype = c_pe.IMAGE_ARM_RUNTIME_FUNCTION_ENTRY
        elif machine == c_pe.IMAGE_FILE_MACHINE.ARM64:
            ctype = c_pe.IMAGE_ARM64_RUNTIME_FUNCTION_ENTRY
        elif machine == c_pe.IMAGE_FILE_MACHINE.ALPHA:
            ctype = c_pe.IMAGE_ALPHA_RUNTIME_FUNCTION_ENTRY
        elif machine == c_pe.IMAGE_FILE_MACHINE.ALPHA64:
            ctype = c_pe.IMAGE_ALPHA64_RUNTIME_FUNCTION_ENTRY
        elif machine in (
            c_pe.IMAGE_FILE_MACHINE.R3000,
            c_pe.IMAGE_FILE_MACHINE.R4000,
            c_pe.IMAGE_FILE_MACHINE.R10000,
            c_pe.IMAGE_FILE_MACHINE.WCEMIPSV2,
            c_pe.IMAGE_FILE_MACHINE.MIPS16,
            c_pe.IMAGE_FILE_MACHINE.MIPSFPU,
            c_pe.IMAGE_FILE_MACHINE.MIPSFPU16,
        ):
            ctype = c_pe.IMAGE_MIPS_RUNTIME_FUNCTION_ENTRY
        else:
            # May be wrong for esoteric architectures, but this is the default
            ctype = c_pe.IMAGE_RUNTIME_FUNCTION_ENTRY

        return ctype[self.size // len(ctype)](self.pe.vfh)



from __future__ import annotations

from functools import cached_property
from io import BytesIO
from typing import TYPE_CHECKING, Any, BinaryIO, TypeAlias

from dissect.util.stream import RangeStream
from dissect.util.ts import from_unix, wintimestamp

from dissect.executable.pe.c_pe import c_pe
from dissect.executable.pe.directory.base import DataDirectory
from dissect.executable.pe.locale_id import CP_TO_NAME, LCID_TO_TAG, TAG_TO_LCID

if TYPE_CHECKING:
    import datetime
    from collections.abc import Iterator

    from dissect.executable.pe.pe import PE


class ResourceDirectory(DataDirectory):
    """The resource directory of a PE file.

    This class provides a higher-level interface to access resources in a PE file. If you wish to access
    the raw resource directory structure, you can traverse it with the ``tree`` attribute.
    """

    def __init__(self, pe: PE, address: int, size: int):
        super().__init__(pe, address, size)
        self.tree = ResourceDirectoryEntry(self, address)

    def __repr__(self) -> str:
        return f"<ResourceDirectory resources={len(self.resources)}>"

    def __len__(self) -> int:
        return len(self.resources)

    def __iter__(self) -> Iterator[Resource]:
        return iter(self.resources)

    def __getitem__(self, idx: int | str | c_pe.RT) -> list[Resource] | None:
        """Get a resource by index or type.

        Args:
            idx: The index of the resource to get, which can be an integer, a string (resource type name),
                 or a member of the ``c_pe.RT`` enum. Integers refer to the index in the resources list,
                 while strings and enum members refer to the resource type.
        """
        if isinstance(idx, int):
            return self.resources[idx]

        if isinstance(idx, (str, c_pe.RT)):
            return self.get(idx)

        raise TypeError(f"ResourceDirectory indices must be int, str or RT enum members, not {type(idx).__name__}")

    def __contains__(self, idx: str | c_pe.RT) -> bool:
        if isinstance(idx, (str, c_pe.RT)):
            return self.get(idx) is not None

        return False

    def get(self, type: int | str | c_pe.RT) -> list[Resource] | None:
        """Get all resources of a specific type.

        Args:
            type: The type of the resource (e.g. ``c_pe.RT.ICON``).
                  Integers can also be used to refer to resource types by their numeric value,
                  or strings to refer to resource types by their name.
        """
        if isinstance(type, int):
            type = c_pe.RT(type)

        if isinstance(type, str) and type in c_pe.RT.__members__:
            type = c_pe.RT[type]

        if (root := self.tree.get(type)) is None:
            return None

        return [Resource(type, name, entry) for name, entry in root.iterdir()]

    def find(self, type: int | str | c_pe.RT, name: str | int) -> Resource | None:
        """Find a specific resource by type and name.

        Args:
            type: The type of the resource (e.g. ``c_pe.RT.ICON``).
                  Integers can also be used to refer to resource types by their numeric value,
                  or strings to refer to resource types by their name.
            name: The name of the resource, which can be a string or an integer ID.
        """
        if isinstance(type, int):
            type = c_pe.RT(type)

        if isinstance(type, str) and type in c_pe.RT.__members__:
            type = c_pe.RT[type]

        if (root := self.tree.get(type)) is None:
            return None

        if isinstance(name, str):
            name = TAG_TO_LCID.get(name, name)

        if isinstance(name, (int, c_pe.RT)) and (entry := root.get(name)) is not None:
            return Resource(type, name, entry)

        raise KeyError(f"Resource with type {type.name!r} and name {name!r} not found")

    @cached_property
    def resources(self) -> list[Resource]:
        """Return a list of all resources."""
        result = []

        for type, type_entry in self.tree.iterdir():
            for name, entry in type_entry.iterdir():
                result.append(Resource(c_pe.RT(type) if isinstance(type, int) else type, name, entry))

        return result

    @property
    def cursor(self) -> list[Resource] | None:
        """Return a list of hardware-dependent cursor resources, if available."""
        return self.get(c_pe.RT.CURSOR)

    @property
    def bitmap(self) -> list[Resource] | None:
        """Return a list of bitmap resources, if available."""
        return self.get(c_pe.RT.BITMAP)

    @property
    def icon(self) -> list[Resource] | None:
        """Return a list of hardware-dependent icon resources, if available."""
        return self.get(c_pe.RT.ICON)

    @property
    def menu(self) -> list[Resource] | None:
        """Return a list of menu resources, if available."""
        return self.get(c_pe.RT.MENU)

    @property
    def dialog(self) -> list[Resource] | None:
        """Return a list of dialog box resources, if available."""
        return self.get(c_pe.RT.DIALOG)

    @property
    def string(self) -> list[Resource] | None:
        """Return a list of string table resources, if available."""
        return self.get(c_pe.RT.STRING)

    def string_table(self, language: str | int | None = None) -> dict[int, str]:
        """Return the parsed string table for a specific language.

        Args:
            language: The language of the string table to return.
                      If ``None``, the first available language is used.
        """
        result = {}

        if isinstance(language, str):
            language = TAG_TO_LCID.get(language, language)

        root = self.tree.get(c_pe.RT.STRING)

        for id, resource in root.iterdir():
            idx = (id - 1) * 16
            entry = next(resource.iterdir())[1] if language is None else resource.get(language)

            with entry.open() as fh:
                while fh.tell() != entry.size:
                    if length := c_pe.USHORT(fh):
                        result[idx] = fh.read(length * 2).decode("utf-16-le")
                    idx += 1

        return result

    @property
    def fontdir(self) -> list[Resource] | None:
        """Return a list of font directory resources, if available."""
        return self.get(c_pe.RT.FONTDIR)

    @property
    def font(self) -> list[Resource] | None:
        """Return a list of font resources, if available."""
        return self.get(c_pe.RT.FONT)

    @property
    def accelerator(self) -> list[Resource] | None:
        """Return a list of accelerator table resources, if available."""
        return self.get(c_pe.RT.ACCELERATOR)

    def accelerator_table(self, language: str | int | None = None) -> dict[int, str]:
        """Return the parsed accelerator table for a specific language.

        Args:
            language: The language of the version information to return.
                      If ``None``, the first available language is used.
        """
        result = []

        if isinstance(language, str):
            language = TAG_TO_LCID.get(language, language)

        for accelerator in self.accelerator or []:
            with accelerator.open(language) as fh:
                last_found = False
                while not last_found:
                    try:
                        key_flags = c_pe.ULONG(fh)
                        modifiers = c_pe.ACCEL_F(key_flags & 0xFFFF)
                        key = c_pe.VK(key_flags >> 16)
                        cmd = c_pe.ULONG(fh)

                        if c_pe.ACCEL_F.LASTKEY in modifiers:
                            last_found = True
                            modifiers &= ~c_pe.ACCEL_F.LASTKEY

                        result.append((key, cmd & 0xFFFF, modifiers))
                    except EOFError:  # noqa: PERF203
                        break

        return result

    @property
    def rcdata(self) -> list[Resource] | None:
        """Return a list of application-defined (raw data) resources, if available."""
        return self.get(c_pe.RT.RCDATA)

    @property
    def message_table(self) -> list[Resource] | None:
        """Return a list of message table resources, if available."""
        return self.get(c_pe.RT.MESSAGETABLE)

    @property
    def group_cursor(self) -> list[Resource] | None:
        """Return a list of hardware-independent group cursor resources, if available."""
        return self.get(c_pe.RT.GROUP_CURSOR)

    @property
    def group_icon(self) -> list[Resource] | None:
        """Return a list of hardware-independent group icon resources, if available."""
        return self.get(c_pe.RT.GROUP_ICON)

    @property
    def version(self) -> list[Resource] | None:
        """Return a list of version resources, if available."""
        return self.get(c_pe.RT.VERSION)

    def vs_version_info(self, language: str | int | None = None) -> dict | None:
        """Return the parsed version information for a specific language.

        Args:
            language: The language of the version information to return.
                      If ``None``, the first available language is used.
        """
        if isinstance(language, str):
            language = TAG_TO_LCID.get(language, language)

        if not (version := self.get(c_pe.RT.VERSION)):
            return None

        if len(version) != 1:
            raise ValueError(f"Expected exactly one version resource, found {len(version)}")

        _Node: TypeAlias = tuple[str, str | bytes | None, list["_Node"]]

        def _parse_lvt(fh: BinaryIO) -> _Node | None:
            start = fh.tell()
            length = c_pe.USHORT(fh)
            if length == 0:
                return None

            value_length = c_pe.USHORT(fh)
            value_type = c_pe.USHORT(fh)

            key = c_pe.WCHAR[None](fh).rstrip("\x00")
            fh.seek(fh.tell() + (-fh.tell() & 3))  # Align to 4 bytes
            value = None
            if value_length:
                value = (
                    fh.read(value_length * 2).decode("utf-16-le").rstrip("\x00")
                    if value_type == 1
                    else fh.read(value_length)
                )
            fh.seek(fh.tell() + (-fh.tell() & 3))  # Align to 4 bytes

            children = []
            while fh.tell() - start < length:
                if (child := _parse_lvt(fh)) is None:
                    break
                children.append(child)

            return key, value, children

        def _build_dict(node: _Node, obj: dict[str, Any]) -> None:
            key, value, children = node

            if key == "VS_VERSION_INFO":
                # Special case for VS_VERSION_INFO, which is a top-level key
                obj[key] = {}
                file_info = c_pe.VS_FIXEDFILEINFO(value.ljust(len(c_pe.VS_FIXEDFILEINFO), b"\x00"))
                if file_info.dwSignature != 0xFEEF04BD:
                    raise ValueError("Invalid VS_FIXEDFILEINFO signature")

                if file_info.dwFileVersionMS or file_info.dwFileVersionLS:
                    obj[key]["FileVersion"] = (
                        f"{file_info.dwFileVersionMS >> 16}.{file_info.dwFileVersionMS & 0xFFFF}.{file_info.dwFileVersionLS >> 16}.{file_info.dwFileVersionLS & 0xFFFF}"  # noqa: E501
                    )

                if file_info.dwProductVersionMS or file_info.dwProductVersionLS:
                    obj[key]["ProductVersion"] = (
                        f"{file_info.dwProductVersionMS >> 16}.{file_info.dwProductVersionMS & 0xFFFF}.{file_info.dwProductVersionLS >> 16}.{file_info.dwProductVersionLS & 0xFFFF}"  # noqa: E501
                    )

                if file_info.dwFileFlags:
                    obj[key]["FileFlags"] = c_pe.VS_FF(file_info.dwFileFlags & file_info.dwFileFlagsMask).name

                if file_info.dwFileOS:
                    upper_file_os = file_info.dwFileOS & 0xFFFF0000
                    lower_file_os = file_info.dwFileOS & 0x0000FFFF
                    obj[key]["FileOS"] = "_".join(
                        c_pe.VOS(x).name for x in filter(None, [upper_file_os, lower_file_os])
                    )

                if file_info.dwFileType:
                    obj[key]["FileType"] = c_pe.VFT(file_info.dwFileType).name

                    if file_info.dwFileSubtype:
                        if file_info.dwFileType == c_pe.VFT.DRV:
                            obj[key]["FileSubtype"] = c_pe.VFT2_DRV(file_info.dwFileSubtype).name
                        elif file_info.dwFileType == c_pe.VFT.FONT:
                            obj[key]["FileSubtype"] = c_pe.VFT2_FONT(file_info.dwFileSubtype).name
                        else:
                            obj[key]["FileSubtype"] = file_info.dwFileSubtype

                if file_info.dwFileDateMS or file_info.dwFileDateLS:
                    obj[key]["FileDate"] = wintimestamp(file_info.dwFileDateMS << 32 | file_info.dwFileDateLS)
            else:
                if value is not None:
                    obj[key] = value
                else:
                    obj[key] = {}

            for child in children:
                _build_dict(child, obj[key])

        result = {}
        with version[0].open(language) as fh:
            _build_dict(_parse_lvt(fh), result)

        if sfi := result.get("VS_VERSION_INFO", {}).get("StringFileInfo", {}):
            keys = list(sfi.keys())
            for key in keys:
                lcid = int(key[:4], 16)
                codepage = int(key[4:], 16)
                sfi[f"{LCID_TO_TAG.get(lcid, str(lcid))}_{CP_TO_NAME.get(codepage, str(codepage))}"] = sfi[key]
                del sfi[key]

        if (vfit := result.get("VS_VERSION_INFO", {}).get("VarFileInfo", {})) and "Translation" in vfit:
            buf = BytesIO(vfit["Translation"])
            tmp = []
            for _ in range(len(vfit["Translation"]) // 4):
                lcid, codepage = c_pe.USHORT[2](buf)
                tmp.append(f"{LCID_TO_TAG.get(lcid, str(lcid))}_{CP_TO_NAME.get(codepage, str(codepage))}")
            vfit["Translation"] = tmp

        return result

    @property
    def dialog_include(self) -> list[Resource] | None:
        """Return a list of dialog include resources, if available."""
        return self.get(c_pe.RT.DLGINCLUDE)

    @property
    def plug_and_play(self) -> list[Resource] | None:
        """Return a list of plug and play resources, if available."""
        return self.get(c_pe.RT.PLUGPLAY)

    @property
    def vxd(self) -> list[Resource] | None:
        """Return a list of VXD resources, if available."""
        return self.get(c_pe.RT.VXD)

    @property
    def animated_cursor(self) -> list[Resource] | None:
        """Return a list of animated cursor resources, if available."""
        return self.get(c_pe.RT.ANICURSOR)

    @property
    def animated_icon(self) -> list[Resource] | None:
        """Return a list of animated icon resources, if available."""
        return self.get(c_pe.RT.ANIICON)

    @property
    def html(self) -> list[Resource] | None:
        """Return a list of HTML resources, if available."""
        return self.get(c_pe.RT.HTML)

    @property
    def manifest(self) -> list[Resource] | None:
        """Return a list of side-by-side assembly manifest resources, if available."""
        return self.get(c_pe.RT.MANIFEST)


class Resource:
    """Higher level representation of a resource in a PE file.

    This class provides a convenient interface to access resource data, abstracting away the details of the
    underlying resource directory structure.

    Args:
        type: The type of the resource (e.g. ``c_pe.RT.ICON``).
        name: The name of the resource, which can be a string or an integer ID.
        entry: The resource directory entry that contains the resource data per language.
    """

    def __init__(self, type: c_pe.RT | str, name: str | int, entry: ResourceDirectoryEntry):
        self.type = type
        self.name = name
        self.entry = entry

    def __repr__(self) -> str:
        return f"<Resource type={self.type!s} name={self.name!r}>"

    def __iter__(self) -> Iterator[Resource]:
        return self.entry.iterdir()

    def __getitem__(self, idx: int | str) -> ResourceDataEntry:
        """Get a specific resource data entry by language.

        Args:
            idx: The index of the resource data entry, which can be an integer (LCID) or a string (language tag).
        """
        if isinstance(idx, str):
            idx = TAG_TO_LCID.get(idx, idx)

        if isinstance(idx, int) and (entry := self.entry.get(idx)) is not None:
            return entry

        raise KeyError(f"Resource with language {idx!r} not found in {self!r}")

    def __contains__(self, idx: str | c_pe.RT) -> bool:
        if isinstance(idx, str):
            idx = TAG_TO_LCID.get(idx, idx)

        if isinstance(idx, int):
            return self.entry.get(idx) is not None

        return False

    def languages(self) -> list[str]:
        """Return a list of languages for this resource."""
        return [LCID_TO_TAG.get(name, str(name)) for name, _ in self.entry.iterdir()]

    def data(self, language: str | int | None = None) -> bytes:
        """Get the resource data for a specific language.

        Args:
            language: The language of the resource to open. If ``None``, the first available resource is opened.
        """
        with self.open(language) as fh:
            return fh.read()

    def open(self, language: str | int | None = None) -> RangeStream:
        """Open the resource data as a stream.

        Args:
            language: The language of the resource to open. If ``None``, the first available resource is opened.
        """
        if language is None:
            return next(self.entry.iterdir())[1].open()

        if isinstance(language, str):
            language = TAG_TO_LCID.get(language, language)

        if isinstance(language, int) and (entry := self.entry.get(language)) is not None:
            return entry.open()

        raise KeyError(f"Resource with language {language!r} not found in {self!r}")


class ResourceEntry:
    """Base class for resource entries in a PE file."""

    def __init__(self, rsrc: ResourceDirectory, address: int):
        self.rsrc = rsrc
        self.address = address

    @cached_property
    def entry(self) -> c_pe.IMAGE_RESOURCE_DATA_ENTRY:
        raise NotImplementedError


class ResourceDataEntry(ResourceEntry):
    """A resource data entry in a PE file."""

    def __repr__(self) -> str:
        return f"<ResourceDataEntry address={self.address:#x} size={self.size} code_page={self.code_page}>"

    @cached_property
    def entry(self) -> c_pe.IMAGE_RESOURCE_DATA_ENTRY:
        """The resource data entry structure."""
        self.rsrc.pe.vfh.seek(self.address)
        return c_pe.IMAGE_RESOURCE_DATA_ENTRY(self.rsrc.pe.vfh)

    @property
    def offset_to_data(self) -> int:
        """The offset to the resource data in the file."""
        return self.entry.OffsetToData

    @property
    def size(self) -> int:
        """The size of the resource data."""
        return self.entry.Size

    @property
    def code_page(self) -> int:
        """The code page of the resource data."""
        return self.entry.CodePage

    @property
    def data(self) -> bytes:
        """The raw resource data."""
        self.rsrc.pe.vfh.seek(self.offset_to_data)
        return self.rsrc.pe.vfh.read(self.size)

    def open(self) -> RangeStream:
        """Open the resource data as a stream."""
        return RangeStream(self.rsrc.pe.vfh, self.offset_to_data, self.size)


class ResourceDirectoryEntry(ResourceEntry):
    """A resource directory entry in a PE file."""

    def __repr__(self) -> str:
        return f"<ResourceDirectoryEntry address={self.address:#x} entries={self.entry.NumberOfNamedEntries + self.entry.NumberOfIdEntries}>"  # noqa: E501

    @cached_property
    def entry(self) -> c_pe.IMAGE_RESOURCE_DIRECTORY:
        """The resource directory entry structure."""
        self.rsrc.pe.vfh.seek(self.address)
        return c_pe.IMAGE_RESOURCE_DIRECTORY(self.rsrc.pe.vfh)

    @property
    def timestamp(self) -> datetime.datetime | None:
        """The timestamp of this resource directory, or ``None`` if the PE file is compiled as reproducible."""
        if self.rsrc.pe.is_reproducible():
            return None
        return from_unix(self.entry.TimeDateStamp)

    def get(self, name: int | str | c_pe.RT) -> ResourceDataEntry | ResourceDirectoryEntry | None:
        """Get a resource entry by name."""
        for id, entry in self.iterdir():
            if id == name:
                return entry
        return None

    def listdir(self) -> dict[int | str, ResourceEntry]:
        """Return a dictionary of the entries in this resource directory."""
        return dict(self.iterdir())

    def iterdir(self) -> Iterator[tuple[int | str, ResourceDataEntry | ResourceDirectoryEntry]]:
        """Iterate over the entries in this resource directory."""
        vfh = self.rsrc.pe.vfh
        offset = self.address + len(c_pe.IMAGE_RESOURCE_DIRECTORY)
        for _ in range(self.entry.NumberOfNamedEntries + self.entry.NumberOfIdEntries):
            vfh.seek(offset)
            entry = c_pe.IMAGE_RESOURCE_DIRECTORY_ENTRY(vfh)

            if entry.NameIsString:
                vfh.seek(self.rsrc.address + entry.NameOffset)
                name = c_pe.IMAGE_RESOURCE_DIR_STRING_U(vfh).NameString
            else:
                name = entry.Id

            if entry.DataIsDirectory:
                obj = ResourceDirectoryEntry(self.rsrc, self.rsrc.address + entry.OffsetToDirectory)
            else:
                obj = ResourceDataEntry(self.rsrc, self.rsrc.address + entry.OffsetToData)

            yield name, obj

            offset += len(c_pe.IMAGE_RESOURCE_DIRECTORY_ENTRY)



from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING

from dissect.executable.pe.c_pe import c_pe
from dissect.executable.pe.directory.base import DataDirectory

if TYPE_CHECKING:
    from collections.abc import Iterator


class TlsDirectory(DataDirectory):
    """The TLS (Thread Local Storage) directory of a PE file."""

    def __repr__(self) -> str:
        return f"<TlsDirectory callbacks={len(self.callbacks)}>"

    def __len__(self) -> int:
        return len(self.callbacks)

    def __iter__(self) -> Iterator[int]:
        return iter(self.callbacks)

    def __getitem__(self, idx: int) -> int:
        return self.callbacks[idx]

    @cached_property
    def header(self) -> c_pe.IMAGE_TLS_DIRECTORY32 | c_pe.IMAGE_TLS_DIRECTORY64:
        """The TLS directory header."""
        self.pe.vfh.seek(self.address)
        ctype = c_pe.IMAGE_TLS_DIRECTORY64 if self.pe.is_64bit() else c_pe.IMAGE_TLS_DIRECTORY32
        return ctype(self.pe.vfh)

    @cached_property
    def callbacks(self) -> list[int]:
        """List of callback addresses."""
        if not self.header.AddressOfCallBacks:
            return []

        self.pe.vfh.seek(self.pe.va_to_rva(self.header.AddressOfCallBacks))
        ctype = c_pe.ULONGLONG if self.pe.is_64bit() else c_pe.ULONG
        try:
            return ctype[None](self.pe.vfh)
        except EOFError:
            return []



from __future__ import annotations

from dataclasses import dataclass
from functools import cached_property
from typing import TYPE_CHECKING

from dissect.executable.pe.c_pe import c_pe
from dissect.executable.pe.directory.base import DataDirectory

if TYPE_CHECKING:
    from collections.abc import Iterator


class BaseRelocationDirectory(DataDirectory):
    """The base relocation directory of a PE file."""

    def __repr__(self) -> str:
        return f"<BaseRelocationDirectory entries={len(self.entries)}>"

    def __len__(self) -> int:
        return len(self.entries)

    def __iter__(self) -> Iterator[BaseRelocation]:
        return iter(self.entries)

    def __getitem__(self, idx: int) -> BaseRelocation:
        return self.entries[idx]

    @cached_property
    def entries(self) -> list[BaseRelocation]:
        """List of base relocation entries."""
        result = []

        offset = self.address
        while offset < self.address + self.size:
            self.pe.vfh.seek(offset)

            block = c_pe._IMAGE_BASE_RELOCATION(self.pe.vfh)
            if block.SizeOfBlock == 0:
                break

            page_rva = block.VirtualAddress

            num_entries = (block.SizeOfBlock - len(c_pe._IMAGE_BASE_RELOCATION)) // len(c_pe.USHORT)
            result.extend(
                BaseRelocation(c_pe.IMAGE_REL_BASED(entry >> 12), page_rva + (entry & 0xFFF))
                for entry in c_pe.USHORT[num_entries](self.pe.vfh)
                if (entry >> 12) != 0  # Skip IMAGE_REL_BASED_ABSOLUTE (0)
            )
            offset += block.SizeOfBlock
            offset += -offset & 3  # Align to 4 bytes

        return result


@dataclass
class BaseRelocation:
    """A single base relocation entry in the base relocation directory."""

    type: c_pe.IMAGE_REL_BASED
    rva: int

    def __repr__(self) -> str:
        return f"<BaseRelocation rva={self.rva:#x} type={self.type.name or self.type.value}>"



from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING

from dissect.executable.pe.c_pe import c_pe
from dissect.executable.pe.directory.base import DataDirectory

if TYPE_CHECKING:
    from collections.abc import Iterator


class SecurityDirectory(DataDirectory):
    """The security directory of a PE file."""

    def __repr__(self) -> str:
        return f"<SecurityDirectory entries={len(self.entries)}>"

    def __len__(self) -> int:
        return len(self.entries)

    def __iter__(self) -> Iterator[Certificate]:
        return iter(self.entries)

    def __getitem__(self, idx: int) -> Certificate:
        return self.entries[idx]

    @cached_property
    def entries(self) -> list[Certificate]:
        """List of certificates in the security directory."""
        result = []

        offset = self.address
        while offset < self.address + self.size:
            # Note: the offset here is a file offset, not an RVA
            self.pe.fh.seek(offset)

            certificate = c_pe.WIN_CERTIFICATE(self.pe.fh)
            if certificate.dwLength == 0:
                break

            result.append(Certificate(certificate))
            offset += certificate.dwLength
            offset += -offset & 7  # Align to 8 bytes

        return result


class Certificate:
    """A single certificate entry in the security directory."""

    def __init__(self, certificate: c_pe.WIN_CERTIFICATE):
        self.certificate = certificate

    def __repr__(self) -> str:
        return f"<Certificate revision={self.revision} type={self.type.name} size={self.certificate.dwLength}>"

    @property
    def revision(self) -> int:
        """The revision of the certificate."""
        return self.certificate.wRevision

    @property
    def type(self) -> c_pe.WIN_CERT_TYPE:
        """The type of the certificate."""
        return self.certificate.wCertificateType

    @property
    def size(self) -> int:
        """The size of the certificate."""
        return self.certificate.dwLength

    @property
    def data(self) -> bytes:
        """The raw data of the certificate."""
        return self.certificate.bCertificate



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from dissect.executable.pe.pe import PE


class DataDirectory:
    """Base class for PE data directories."""

    def __init__(self, pe: PE, address: int, size: int):
        self.pe = pe
        self.address = address
        self.size = size

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} address={self.address:#x} size={self.size}>"



from __future__ import annotations

from dissect.executable.elf.elf import (
    ELF,
    Section,
    SectionTable,
    Segment,
    SegmentTable,
    StringTable,
    Symbol,
    SymbolTable,
)

__all__ = [
    "ELF",
    "Section",
    "SectionTable",
    "Segment",
    "SegmentTable",
    "StringTable",
    "Symbol",
    "SymbolTable",
]



from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING

from dissect.util.ts import from_unix

from dissect.executable.pe.c_pe import c_pe
from dissect.executable.pe.directory.base import DataDirectory

if TYPE_CHECKING:
    import datetime


class BoundImportDirectory(DataDirectory):
    """The bound import directory of a PE file."""

    def __repr__(self) -> str:
        return f"<BoundImportDirectory modules={len(self.modules)}>"

    def __len__(self) -> int:
        return len(self.modules)

    def __getitem__(self, idx: str | int) -> BoundImportModule:
        if isinstance(idx, int):
            return self.modules[idx]
        if isinstance(idx, str):
            if idx not in self._by_name:
                raise KeyError(f"Bound import module {idx!r} not found")
            return self._by_name[idx]
        raise TypeError(f"BoundImportDirectory indices must be str or int, not {type(idx).__name__}")

    def __contains__(self, name: str) -> bool:
        if isinstance(name, str):
            return name in self._by_name
        return False

    @cached_property
    def modules(self) -> list[BoundImportModule]:
        """List of bound imported modules."""
        result = []

        self.pe.vfh.seek(self.address)
        while self.pe.vfh.tell() < self.address + self.size:
            descriptor = c_pe.IMAGE_BOUND_IMPORT_DESCRIPTOR(self.pe.vfh)
            if not descriptor:
                break

            forwarders = []
            for _ in range(descriptor.NumberOfModuleForwarderRefs):
                forwarder = c_pe.IMAGE_BOUND_FORWARDER_REF(self.pe.vfh)
                if not forwarder:
                    break

                forwarders.append(BoundImportForwardReference(self, forwarder))

            result.append(BoundImportModule(self, descriptor, forwarders))

        return result

    @cached_property
    def _by_name(self) -> dict[str, BoundImportModule]:
        """A mapping of module names to their :class:`DelayImportModule`."""
        return {module.name: module for module in self.modules}


class BoundImportModule:
    """A module bound imported by a PE file, containing its functions."""

    def __init__(
        self,
        directory: BoundImportDirectory,
        descriptor: c_pe.IMAGE_BOUND_IMPORT_DESCRIPTOR,
        forwarders: list[BoundImportForwardReference],
    ):
        self.directory = directory
        self.descriptor = descriptor
        self.forwarders = forwarders

    @property
    def timestamp(self) -> datetime.datetime | None:
        """The timestamp of this bound import module, or ``None`` if the PE file is compiled as reproducible."""
        if self.directory.pe.is_reproducible():
            return None
        return from_unix(self.descriptor.TimeDateStamp)

    @property
    def name(self) -> str:
        self.directory.pe.vfh.seek(self.directory.address + self.descriptor.OffsetModuleName)
        return c_pe.CHAR[None](self.directory.pe.vfh).decode()


class BoundImportForwardReference:
    """A forward reference in a bound import module."""

    def __init__(
        self,
        directory: BoundImportDirectory,
        descriptor: c_pe.IMAGE_BOUND_FORWARDER_REF,
    ):
        self.directory = directory
        self.descriptor = descriptor

    @property
    def timestamp(self) -> datetime.datetime | None:
        """The timestamp of this bound import module, or ``None`` if the PE file is compiled as reproducible."""
        if self.directory.pe.is_reproducible():
            return None
        return from_unix(self.descriptor.TimeDateStamp)

    @property
    def name(self) -> str:
        self.directory.pe.vfh.seek(self.directory.address + self.descriptor.OffsetModuleName)
        return c_pe.CHAR[None](self.directory.pe.vfh).decode()



from __future__ import annotations

from functools import cached_property
from typing import BinaryIO

from dissect.util.stream import RangeStream

from dissect.executable.pe.c_pe import c_pe
from dissect.executable.pe.directory.base import DataDirectory


class ComDescriptorDirectory(DataDirectory):
    """The COM descriptor directory of a PE file.

    References:
        - https://www.codeproject.com/Articles/12585/The-NET-File-Format
    """

    @cached_property
    def descriptor(self) -> c_pe.IMAGE_COR20_HEADER:
        """The CLR 2.0 header descriptor."""
        self.pe.vfh.seek(self.address)
        return c_pe.IMAGE_COR20_HEADER(self.pe.vfh)

    @cached_property
    def metadata(self) -> ComMetadata:
        """The COM metadata directory."""
        return ComMetadata(self.pe, self.descriptor.MetaData.VirtualAddress, self.descriptor.MetaData.Size)


class ComMetadata(DataDirectory):
    """The COM metadata directory of the COM descriptor."""

    @cached_property
    def metadata(self) -> c_pe.IMAGE_COR20_METADATA:
        """The CLR 2.0 metadata descriptor."""
        self.pe.vfh.seek(self.address)
        return c_pe.IMAGE_COR20_METADATA(self.pe.vfh)

    @property
    def version(self) -> str:
        """The version as defined in the metadata."""
        return self.metadata.Version.decode().strip("\x00")

    @cached_property
    def streams(self) -> list[ComStream]:
        """A list of streams defined in the metadata."""
        result = []

        offset = self.address + len(self.metadata)
        for _ in range(self.metadata.NumberOfStreams):
            self.pe.vfh.seek(offset)
            header = c_pe.IMAGE_COR20_STREAM_HEADER(self.pe.vfh)

            result.append(ComStream(self, header.Offset, header.Size, header.Name.decode()))

            offset += len(header)
            offset += -offset & 3  # Align to 4 bytes

        return result


class ComStream:
    """A stream in the COM metadata."""

    def __init__(self, metadata: ComMetadata, offset: int, size: int, name: str):
        self.metadata = metadata
        self.offset = offset
        self.size = size
        self.name = name

    def __repr__(self) -> str:
        return f"<ComStream offset={self.offset:#x} size={self.size:#x} name={self.name!r}>"

    @property
    def data(self) -> bytes:
        """The data of the stream."""
        self.metadata.pe.vfh.seek(self.metadata.address + self.offset)
        return self.metadata.pe.vfh.read(self.size)

    def open(self) -> BinaryIO:
        """Open the stream for reading."""
        return RangeStream(self.metadata.pe.vfh, self.metadata.address + self.offset, self.size)



from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING

from dissect.executable.pe.c_pe import c_pe
from dissect.executable.pe.directory.base import DataDirectory

if TYPE_CHECKING:
    from collections.abc import Iterator


class IatDirectory(DataDirectory):
    """The import address table (IAT) directory of a PE file."""

    def __repr__(self) -> str:
        return f"<IatDirectory entries={len(self.entries)}>"

    def __len__(self) -> int:
        return len(self.entries)

    def __iter__(self) -> Iterator[int]:
        return iter(self.entries)

    def __getitem__(self, idx: int) -> int:
        return self.entries[idx]

    @cached_property
    def entries(self) -> list[int]:
        """List of addresses in the import address table."""
        self.pe.vfh.seek(self.address)
        ctype = c_pe.ULONGLONG if self.pe.is_64bit() else c_pe.ULONG
        return ctype[self.size // len(ctype)](self.pe.vfh)



from __future__ import annotations

from dissect.executable.pe.directory.base import DataDirectory
from dissect.executable.pe.directory.basereloc import BaseRelocationDirectory
from dissect.executable.pe.directory.bound_import import BoundImportDirectory
from dissect.executable.pe.directory.com_descriptor import ComDescriptorDirectory
from dissect.executable.pe.directory.debug import DebugDirectory
from dissect.executable.pe.directory.delay_import import DelayImportDirectory
from dissect.executable.pe.directory.exception import ExceptionDirectory
from dissect.executable.pe.directory.export import ExportDirectory
from dissect.executable.pe.directory.iat import IatDirectory
from dissect.executable.pe.directory.imports import ImportDirectory, ImportFunction, ImportModule
from dissect.executable.pe.directory.load_config import LoadConfigDirectory
from dissect.executable.pe.directory.resource import (
    ResourceDataEntry,
    ResourceDirectory,
    ResourceDirectoryEntry,
    ResourceEntry,
)
from dissect.executable.pe.directory.security import SecurityDirectory
from dissect.executable.pe.directory.tls import TlsDirectory

__all__ = [
    "BaseRelocationDirectory",
    "BoundImportDirectory",
    "ComDescriptorDirectory",
    "DataDirectory",
    "DebugDirectory",
    "DelayImportDirectory",
    "ExceptionDirectory",
    "ExportDirectory",
    "IatDirectory",
    "ImportDirectory",
    "ImportFunction",
    "ImportModule",
    "LoadConfigDirectory",
    "ResourceDataEntry",
    "ResourceDirectory",
    "ResourceDirectoryEntry",
    "ResourceEntry",
    "SecurityDirectory",
    "TlsDirectory",
]
